
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	Gerbil
//
//	Copyright (c) 2001, Bruce Moreland.  All rights reserved.
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	This file is part of the Gerbil chess program project.
//
//	Gerbil is free software; you can redistribute it and/or modify it under
//	the terms of the GNU General Public License as published by the Free
//	Software Foundation; either version 2 of the License, or (at your option)
//	any later version.
//
//	Gerbil is distributed in the hope that it will be useful, but WITHOUT ANY
//	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
//	FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
//	details.
//
//	You should have received a copy of the GNU General Public License along
//	with Gerbil; if not, write to the Free Software Foundation, Inc.,
//	59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#include "engine.hpp"
#include "board.hpp"
#include "hash.hpp"
#include "moves.hpp"
#include "genmagic.hpp"
#include "deprec.hpp"

#include <ctype.h>

#ifdef	_DEBUG
static char const s_aszModule[] = __FILE__;
#endif

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
#define	isqA1	0
#define	isqB1	1
#define	isqC1	2
#define	isqD1	3
#define	isqE1	4
#define	isqF1	5
#define	isqG1	6
#define	isqH1	7
#define	isqA2	16
#define	isqB2	17
#define	isqC2	18
#define	isqD2	19
#define	isqE2	20
#define	isqF2	21
#define	isqG2	22
#define	isqH2	23
#define	isqA3	32
#define	isqB3	33
#define	isqC3	34
#define	isqD3	35
#define	isqE3	36
#define	isqF3	37
#define	isqG3	38
#define	isqH3	39
#define	isqA4	48
#define	isqB4	49
#define	isqC4	50
#define	isqD4	51
#define	isqE4	52
#define	isqF4	53
#define	isqG4	54
#define	isqH4	55
#define	isqA5	64
#define	isqB5	65
#define	isqC5	66
#define	isqD5	67
#define	isqE5	68
#define	isqF5	69
#define	isqG5	70
#define	isqH5	71
#define	isqA6	80
#define	isqB6	81
#define	isqC6	82
#define	isqD6	83
#define	isqE6	84
#define	isqF6	85
#define	isqG6	86
#define	isqH6	87
#define	isqA7	96
#define	isqB7	97
#define	isqC7	98
#define	isqD7	99
#define	isqE7	100
#define	isqF7	101
#define	isqG7	102
#define	isqH7	103
#define	isqA8	112
#define	isqB8	113
#define	isqC8	114
#define	isqD8	115
#define	isqE8	116
#define	isqF8	117
#define	isqG8	118
#define	isqH8	119

//	Candidate move flags.  The move generator marks each move it generates
//	with zero or more of these.

#define	cmfNONE		0x00000000	// A normal move.
#define	cmfPR_MASK	0x0000000F	// If a move promotes, the "pc" it promotes
								//  to will be stuck in these bits.
#define	cmfSET_ENP	0x00010000	// e2-e4 will set this flag.
#define	cmfMAKE_ENP	0x00020000	// An en-passant capture will set this flag.
#define	cmfCASTLE	0x00040000	// O-O will set this flag.
#define	cmfCAPTURE	0x00080000	// Any capture (including en-passant) will set
								//  this flag.

//	PROM_Q,		PROM_R,		PROM_B,		PROM_N
static char promPc[] = {'Q','R','B','N'};


//	Inits context ("pcon") from FEN string.  Returns FALSE if the position
//	is illegal.

//	"VUndoMove" makes a few assumptions about how this routine works.  If you
//	go messing with the "pcon->gc" info in here, check that routine to make
//	sure you don't break it.

//	This routine is extremely important, because it sets up the position and
//	makes sure the program is ready to go.

bool SetBoardFromFEN(PCON pcon, char const * szFen)
{
	int	i;
	int	argcKing[coMAX];
	int	argvalPc[coMAX];
	int	argvalPn[coMAX];
	PSTE	pste = pcon->argste;

	strcpy(pcon->gc.aszFen, szFen);	// Remember FEN.
	pcon->argcpi[W] = 0;		// Number of pieces per side.
	pcon->argcpi[B] = 0;
	for (i = 0; i < csqMAX; i++) {	// Clear board.
		pcon->argsq[i].ppi = NULL;
		pcon->argsq[i].isq = i;
	}
	pste->hashkPc = 0;		// Going to XOR the all pieces/pawns into this.
	pste->hashkPn = 0;		// Going to XOR the pawns only into this.
	argvalPc[W] = argvalPc[B] = 0;
	argvalPn[W] = argvalPn[B] = 0;
	//
	//	Fill in pcon from FEN string.  Part one of the FEN is piece locations,
	//	so this fills in the board and the piece list.
	//
	int fil = filA;
	int rnk = rnk8;
	argcKing[coWHITE] = argcKing[coBLACK] = 0;
	for (;; szFen++) {
		if (*szFen == ' ') {
			szFen++;
			break;
		}
		switch (*szFen) {
			PPI	ppi;
			int	pc;
			int	co;

		default:			// Includes '\0' (end of input).
			return false;
		case 'p':
			pc = pcPAWN;
			co = coBLACK;
lblSetPawn:	if ((rnk == rnk1) ||			// Pawn in a stupid place.
				(rnk == rnk8))
				return false;
lblSet:		if (pcon->argcpi[co] >= cpiMAX)	// Too many pieces of this color.
				return false;
			if (rnk < rnk1)					// Ran off the board.
				return false;
			if (fil > filH)
				return false;
			ppi = &pcon->argpi[co][pcon->argcpi[co]++];
			//
			//	Patch king so it's at the front of the list.
			//
			Assert(ipiKING == 0);	// This won't work if it's not zero.
			if ((pc == pcKING) && (pcon->argcpi[co] > 1)) {
				PI pi = pcon->argpi[co][ipiKING];

				*ppi = pi;
				pcon->argsq[ppi->isq].ppi = ppi;
				ppi = &pcon->argpi[co][ipiKING];
			}
			//	Create new piece.
			//
			ppi->pc = pc;
			ppi->co = co;
			ppi->fDead = false;
			ppi->isq = IsqFromRnkFil(rnk, fil);
			argvalPc[co] += s_argvalPcOnly[pc];
			argvalPn[co] += s_argvalPnOnly[pc];
			pcon->argsq[ppi->isq].ppi = ppi;
//			rehash( pste->hashkPc, pc, co, ppi->isq );
//			if (pc == pcPAWN)
//				rehash( pste->hashkPn, pc, co, ppi->isq );
			fil++;
			break;
		case 'P':
			pc = pcPAWN;
			co = coWHITE;
			goto lblSetPawn;
		case 'n':
			pc = pcKNIGHT;
			co = coBLACK;
			goto lblSet;
		case 'N':
			pc = pcKNIGHT;
			co = coWHITE;
			goto lblSet;
		case 'b':
			pc = pcBISHOP;
			co = coBLACK;
			goto lblSet;
		case 'B':
			pc = pcBISHOP;
			co = coWHITE;
			goto lblSet;
		case 'r':
			pc = pcROOK;
			co = coBLACK;
			goto lblSet;
		case 'R':
			pc = pcROOK;
			co = coWHITE;
			goto lblSet;
		case 'q':
			pc = pcQUEEN;
			co = coBLACK;
			goto lblSet;
		case 'Q':
			pc = pcQUEEN;
			co = coWHITE;
			goto lblSet;
		case 'k':
			argcKing[coBLACK]++;
			pc = pcKING;
			co = coBLACK;
			goto lblSet;
		case 'K':
			argcKing[coWHITE]++;
			pc = pcKING;
			co = coWHITE;
			goto lblSet;
		case '/':
			fil = filA;
			rnk--;
			break;
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
			fil += *szFen - '0';
			break;
		}
	}
	if ((argcKing[coWHITE] != 1) || (argcKing[coBLACK] != 1))
		return false;		// Too many or not enough kings.  I could check
							//  for other cases of too many pieces ( > 9 Q's),
							//  but that doesn't seem necessary yet.
							// I'm also allowing > 8 pawns.
	//
	//	At this point I'm sitting beyond the first ' '.  Collect color, which
	//	must be either 'w' or 'b'.
	//
	switch (*szFen++) {
	case 'w':
		pste->side = W;
		pcon->gc.coStart = W;
		break;
	case 'b':
		pste->side = B;
		pcon->gc.coStart = B;
		pste->hashkPc = HashkSwitch(pste->hashkPc);
		break;
	default:			// Includes '\0' (end of input).
		return false;
	}
    pste->xside = pste->side ^ 1;
    int side = pste->side;
	pste->valPcUs = argvalPc[side];
	pste->valPcThem = argvalPc[side ^ 1];
	pste->valPnUs = argvalPn[side];
	pste->valPnThem = argvalPn[side ^ 1];
	//
	//	I'm sitting on the ' ' beyond color.  Force it, then get castling
	//	flags.
	//
	if (*szFen++ != ' ')
		return false;
	pste->castle = cfNONE;
	if (*szFen == '-') {
		szFen++;
		if (*szFen++ != ' ')
			return false;
	} else {
		for (;;) {
			if (*szFen == ' ') {
				szFen++;
				break;
			}
			switch (*szFen++) {
			default:			// Includes '\0' (end of input).
				return false;
			case 'k':
				pste->castle |= cfE8G8;
				break;
			case 'q':
				pste->castle |= cfE8C8;
				break;
			case 'K':
				pste->castle |= cfE1G1;
				break;
			case 'Q':
				pste->castle |= cfE1C1;
				break;
			}
		}
		//	There are a lot of FEN's with bogus castling flags.  Rather than
		//	be anal about this I'm going to do some quick checks and eliminate
		//	bogus flags.
		//
		if ((pcon->argsq[isqE1].ppi == NULL) ||
			(pcon->argsq[isqE1].ppi->pc != pcKING) ||
			(pcon->argsq[isqE1].ppi->co != coWHITE))
			pste->castle &= ~(cfE1G1 | cfE1C1);
		if ((pcon->argsq[isqA1].ppi == NULL) ||
			(pcon->argsq[isqA1].ppi->pc != pcROOK) ||
			(pcon->argsq[isqA1].ppi->co != coWHITE))
			pste->castle &= ~cfE1C1;
		if ((pcon->argsq[isqH1].ppi == NULL) ||
			(pcon->argsq[isqH1].ppi->pc != pcROOK) ||
			(pcon->argsq[isqH1].ppi->co != coWHITE))
			pste->castle &= ~cfE1G1;
		if ((pcon->argsq[isqE8].ppi == NULL) ||
			(pcon->argsq[isqE8].ppi->pc != pcKING) ||
			(pcon->argsq[isqE8].ppi->co != coBLACK))
			pste->castle &= ~(cfE8G8 | cfE8C8);
		if ((pcon->argsq[isqA8].ppi == NULL) ||
			(pcon->argsq[isqA8].ppi->pc != pcROOK) ||
			(pcon->argsq[isqA8].ppi->co != coBLACK))
			pste->castle &= ~cfE8C8;
		if ((pcon->argsq[isqH8].ppi == NULL) ||
			(pcon->argsq[isqH8].ppi->pc != pcROOK) ||
			(pcon->argsq[isqH8].ppi->co != coBLACK))
			pste->castle &= ~cfE8G8;
	}
	//	I'm sitting after the space after castling flags.  Time to process
	//	en-passant square.
	//
	//	I'll check to see if the square is in the range a1..h8, and if not
	//	I'll return FALSE.
	//
	//	If the square seems okay, I'll check to see if a capture might be
	//	possible, because sometimes people put bogus EnP squares in their
	//	FEN string.  If the capture looks impossible, I'll null-out the
	//	square.  "Might be possible" means that the target square is empty,
	//	there is a pawn that could capture, and a pawn that can be captured.
	//
	//	I'm not going to check to see if an en-passant capture is strictly
	//	legal.
	//
	pste->isqEnP = isqNIL;
	if (*szFen == '-')
		szFen++;
	else {
		int	isqCapture;
		int	isqEnP;

		if ((*szFen < 'a') || (*szFen > 'h'))
			return false;
		if ((*(szFen + 1) < '1') || (*(szFen + 1) > '8'))
			return false;
		//
		//	If here, square is in the range a1..h8.
		//
		isqEnP = isqCapture = isqNIL;
		if (side == W) {
			if (*(szFen + 1) == '6') {
				isqEnP = *szFen - 'a' + (*(szFen + 1) - '1') * filIBD;
				isqCapture = isqEnP - filIBD;
			}
		} else {
			if (*(szFen + 1) == '3') {
				isqEnP = *szFen - 'a' + (*(szFen + 1) - '1') * filIBD;
				isqCapture = isqEnP + filIBD;
			}
		}
		//	"isqCapture" is the square of the captured pawn, given that
		//	"isqEnP" is where the capturing pawn ends up.  For instance, in
		//	the en-passant capture "exd6", the "isqEnP" square is e6, since
		//	that is where the capturing pawn goes, and "isqCapture" is d5,
		//	since that's the square the pawn is removed from.
		//
		if (isqEnP != isqNIL) {
			PPI	ppiCapture = pcon->argsq[isqCapture].ppi;

			//	Check to see if there is an appropriate pawn on the capture
			//	square.
			//
			//	If this is not true, this is a bogus en-passant square, and
			//	I'm going to null it out before it causes some sort of
			//	problem.
			//
			//	The same conditions apply during the search.
			//
			if ((ppiCapture == NULL) || (ppiCapture->pc != pcPAWN) ||
				(ppiCapture->co == side))
				isqEnP = isqNIL;
		}
		pste->isqEnP = isqEnP;
		szFen += 2;
	}
	if (*szFen++ != ' ')
		return false;
	//
	//	Get fifty-move counter.
	//
	pste->plyFifty = 0;
	for (;;) {
		if (!isdigit(*szFen))
			break;
		pste->plyFifty *= 10;
		pste->plyFifty += *szFen++ - '0';
	}
	if (*szFen++ != ' ')
		return false;
	//
	//	Get start move counter.
	//
	pcon->gc.movStart = 0;
	for (;;) {
		if (!isdigit(*szFen))
			break;
		pcon->gc.movStart *= 10;
		pcon->gc.movStart += *szFen++ - '0';
	}
	if (pcon->gc.movStart < 1)	// Start move must be 1 or more.
		return false;
	//
	//	Look for extraneous crap at the end of the FEN.
	//
	while (*szFen == ' ')
		szFen++;
	if (*szFen != '\0')
		return false;
	//
	//	Sets number of moves in this game (zero), and record the current
	//	position's hash key for rep checking.
	//
	pcon->gc.ccm = 0;				// Zero moves made so far.
	//
	//	"pcon->argste" contains some precomputed stuff, and this sets that
	//	up properly.
	//
	VFixSte(pcon);
    convMailboxtoBB(pcon, pste);
    pste->mlist_start = pcon->move_list;
	//
	//	The position at the start of the game has occurred one time, so twice
	//	more is a draw.  This needs to be recorded in case it needs to be
	//	claimed.
	//
	pcon->gc.arghashk[0] = pste->hashkPc;
	pcon->gc.arghashkFull[0] = HashkFull(pcon, pste);
	return true;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This routine takes a "SAN" record and produces a SAN move output string
//	(such as "e4", or "dxe6+", or "Nbd7", or "Qxg7#").

//	The SAN element already has all the information needed to do this, so it's
//	a very simple process.

void SanToSz(const SAN *psan, char * sz)
{
	if (psan->cmf & cmfCASTLE) {
		if ((psan->isqTo == G1) || (psan->isqTo == G8))
			sz += sprintf(sz, "O-O");
		else
			sz += sprintf(sz, "O-O-O");
	} else {
		if (psan->pc == WP || psan->pc == BP )
			*sz++ = char(File(psan->isqFrom) + 'a');
		else {
			*sz++ = pc_char_map[psan->pc|1];
			if (psan->sanf & sanfRNK_AMBIG)
				*sz++ = char(File(psan->isqFrom) + 'a');
			if (psan->sanf & sanfFIL_AMBIG)
				*sz++ = char(Rank(psan->isqFrom) + '1');
		}
		if (psan->cmf & cmfCAPTURE) {
			*sz++ = 'x';
			*sz++ = char(File(psan->isqTo) + 'a');
			*sz++ = char(Rank(psan->isqTo) + '1');
		} else if ((psan->pc|1) == WP)
			*sz++ = char(Rank(psan->isqTo) + '1');
		else {
			*sz++ = char(File(psan->isqTo) + 'a');
			*sz++ = char(Rank(psan->isqTo) + '1');
		}
		if (psan->cmf & cmfPR_MASK) {
			*sz++ = '=';
			*sz++ = promPc[(psan->cmf & cmfPR_MASK)-1];
		}
	}
	if (psan->sanf & sanfMATE)
		*sz++ = '#';
	else if (psan->sanf & sanfCHECK)
		*sz++ = '+';
	*sz = '\0';
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This takes a CM and converts it to a SAN record, which can be used by
//	"VSanToSz" to make something pretty to display.

//	This is a terrible process, in large part because of disambiguation.
//	Disambiguation would be extremely easy, if all I had to do is disambiguate
//	between moves produced by the move generator.  But some of these moves are
//	not actually legal, so I have to delete those.

//	Also, part of the process includes appending "+", or "#" if necessary,
//	which especially in the latter case is annoying.

//	So this routine figures out if a given move is ambiguous, and if so
//	records exactly how, and it also figures out if the move is check or mate.

//	WARNING! This function won't work if the moves aren't generated for this
//	position, and these moves are pseudo-legal, not fully legal.

void CmToSan(PCON pcon, PSTE pste, const CM *pcmSrc, PSAN psan)
{
	//	Easy stuff first.
	//
    U32 move = pcmSrc->m;
	// Decompress the move
	int from	= From(move);
	int to		= To(move);
    int pc      = Pc(move);

	psan->pc = pc;
	psan->isqFrom = from;
	psan->isqTo = to;
    psan->cmf = 0;
    if( Prom(move) )
        to=to;
    psan->m   = move;
	psan->sanf = sanfNONE;
    if( Capt(move) ) {
        psan->cmf |= cmfCAPTURE;
    }
    if( (pc|1) == (PROM|1)) {
        psan->pc = PAWN | (pc & 1);
        psan->cmf |= Prom(move) + 1;
    }
    if( pc == B_CASTLE) {
    	psan->pc = BK;
        psan->cmf |= cmfCASTLE;
    }
    if( pc == W_CASTLE) {
    	psan->pc = WK;
        psan->cmf |= cmfCASTLE;
    }

	//
	//	Disambiguate.
	//
	bool fAmbiguous = false;
    foreachmove(pste,pcm) {
		if ( (To(pcm->m) == To(move)) && From(pcm->m) != From(move) &&
            Pc(pcm->m) == Pc(move)) {
			bMove(pcon, pste, pcm->m);
			if (Rank(From(pcm->m)) == Rank(From(move)))
				psan->sanf |= sanfRNK_AMBIG;
			if (File(From(pcm->m)) == File(From(move)))
				psan->sanf |= sanfFIL_AMBIG;
			fAmbiguous = true;
			bUndoMove(pcon, pste, pcm->m);
		}
    }
	if ((fAmbiguous) && (!(psan->sanf & (sanfRNK_AMBIG | sanfFIL_AMBIG))))
		psan->sanf |= sanfRNK_AMBIG;	// "Nbd2" rather than "N1d2".
	//
	//	The gross disambiguation has already been done.  Now I'm going to
	//	execute the move and see if it is check or mate.
	//
	bMove(pcon, pste, pcmSrc->m);
//	if ( inCheck(pcon, pste->side ^ 1) ) {
	if ( pste[1].checkf ) {
		psan->sanf |= sanfCHECK;
		bGenEvasions(pcon, pste + 1);
        if( legalMoveCount( pste + 1 ) == 0 )
			psan->sanf |= sanfMATE;
	}
	bUndoMove(pcon, pste, pcmSrc->m);
	//
	//	That was painful.
}

//	This converts a CM to "e2e4" (or "e7e8q") format.  I'm writing this with
//	Winboard in mind, and that is a format that Winboard will accept.  It is
//	also very easy to generate.
void CmToSz(const CM *pcm, char * sz)
{
    U32 m = pcm->m;
	sz += sprintf(sz, "%c%c%c%c",
		File(From(m)) + 'a',
		Rank(From(m)) + '1',
		File(To(m)) + 'a',
		Rank(To(m)) + '1');
	if (Pc(m) == W_PROM || Pc(m) == B_PROM )
		sz += sprintf(sz, "%c", promPc[Prom(m)]);
}


