
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	Gerbil
//
//	Copyright (c) 2001, Bruce Moreland.  All rights reserved.
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	This file is part of the Gerbil chess program project.
//
//	Gerbil is free software; you can redistribute it and/or modify it under
//	the terms of the GNU General Public License as published by the Free
//	Software Foundation; either version 2 of the License, or (at your option)
//	any later version.
//
//	Gerbil is distributed in the hope that it will be useful, but WITHOUT ANY
//	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
//	FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
//	details.
//
//	You should have received a copy of the GNU General Public License along
//	with Gerbil; if not, write to the Free Software Foundation, Inc.,
//	59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#include "engine.hpp"
#include "eval.hpp"
#include "moves.hpp"
#include "genmagic.hpp"

#ifdef _DEBUG
static char const s_aszModule[] = __FILE__;
#endif


const U64 fileAMask	= 0x0101010101010101ULL;
const U64 fileHMask = 0x8080808080808080ULL;
//#define fileAMask	0x0101010101010101ULL
//#define fileHMask   0x8080808080808080ULL

const int promotionSqrFileA[2] = {A1,A8};
const int promotionSqrFileH[2] = {H1,H8};

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

const unsigned char castle_mask[2] = { 12, 3 };

// I don't want to use one array per color for positional values
// this array precomputes rank swap
int const remapSq[2][64] = {
    {
//  a   b   c   d   e   f   g   h
	0,	1,	2,	3,	4,	5,	6,	7,
	8,	9,	10,	11,	12,	13,	14,	15,
	16,	17,	18,	19,	20,	21,	22,	23,
	24,	25,	26,	27,	28,	29,	30,	31,
	32,	33,	34,	35,	36,	37,	38,	39,
	40,	41,	42,	43,	44,	45,	46,	47,
	48,	49,	50,	51,	52,	53,	54,	55,
	56,	57,	58,	59,	60,	61,	62,	63
    },{
	56,	57,	58,	59,	60,	61,	62,	63,
	48,	49,	50,	51,	52,	53,	54,	55,
	40,	41,	42,	43,	44,	45,	46,	47,
	32,	33,	34,	35,	36,	37,	38,	39,
	24,	25,	26,	27,	28,	29,	30,	31,
	16,	17,	18,	19,	20,	21,	22,	23,
	8,	9,	10,	11,	12,	13,	14,	15,
	0,	1,	2,	3,	4,	5,	6,	7
    }
};

//
//	These are piece-square tables.  The award bonuses (or penalties) in
//	centipawns if a piece is sitting on a particular square
//

/*
const int KING		= 0;    => 0
const int QUEEN		= 2;    => 1
const int ROOK		= 4;    => 2
const int BISHOP	= 6;    => 3
const int KNIGHT	= 8;    => 4
const int PAWN		= 10;   => 5
*/
static int	const staticPosVal[6+2][64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-7,	-15,-15,-15,-15,-15,-15,-7,
	-5,	-5,	-12,-12,-12,-12,-5,	-5,
	 3,	 5,	  6,-12, -9,-12, 8,	 6,
	//
	//	White queen.
	//
//  a   b   c   d   e   f   g   h
	-2,	-2,	0,	0,	0,	0,	0,	0,
	-2,	-2,	0,	0,	0,	0,	0,	0,
	-2,	-2,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	2,	2,	0,	0,	0,
	0,	0,	0,	2,	2,	0,	0,	0,
	0,	1,	1,	1,	1,	0,	0,	0,
	0,	0,	1,	1,	1,	0,	0,	0,
	-2,	-2,	-2,	0,	0,	-2,	-2,	-2,
	//
	//	White rook.
	//
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,
	8,	8,	8,	8,	8,	8,	8,	8,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-4,	0,	0,	0,	0,	0,	0,	-4,
	-4,	0,	3,	6,	6,	3,	0,	-4,
	//
	//	White bishop.
	//
//  a   b   c   d   e   f   g   h
	5,	5,	5,	5,	5,	5,	5,	5,
	8,	8,	8,  8,  8,	8,	8,	8,
	8,	8,	8,  8,  8,	8,	8,	8,
	8,	9,	8,  8,  8,	8,  9,	8,
	8,	8,  9,  8,  8,  9,	8,	8,
	6,	8,	8,  8,  8,	8,	8,	6,
	4,  7,	5,  7,  7,	5,  7,	4,
	0, -4,	0,	0,	0,	0, -4,	0,
	//
	//	White knight.
	//
//  a   b   c   d   e   f   g   h
	-9,-5,-5, -5, -5, -5, -5, -9,
	-5,	0,  8,  8,  8,  8,	0, -5,
	-5,	5,  9,  9,  9,  9,	5, -5,
	-5,	7,	7,  9,  9,	7,	7, -5,
	-5,	8,	5,	6,	6,	5,	8, -5,
	-5,	5,  8,	5,	5,  8,	5, -5,
	-5,	0,  5,  7,  7,  5,	0, -5,
	-9,-2, -2, -2, -2, -2, -2, -9,
	//
	//	White pawn.
	//
// TODO:not up to date =============================================================
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,  // a8..h8
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0, 15, 15,	0,	0,	0,
	0,	0,	4,	7,	7,	0,	0,	0,
	0,	0,	5,  7,  7,	4,	0,	0,
	2,	2,	1,	5,	5,	0,	2,	2,
	5,	5,	2,	0,	0,	5,	5,	5,
	0,	0,	0,	0,	0,	0,	0,	0,  // a1..h1
};

//
//

static int	const tempoVal[6+2][64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  4,  0,  0,  0,  4,  0,
	//
	//	White queen.
	//
//  a   b   c   d   e   f   g   h
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	//
	//	White rook.
	//
//  a   b   c   d   e   f   g   h
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	0,  4,  4,  4,  4,  4,  4,  0,
	0,  0,  4,  4,  4,  4,  0,  0,
	//
	//	White bishop.
	//
//  a   b   c   d   e   f   g   h
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  0,  4,  4,  0,  4,  4,
	//
	//	White knight.
	//
//  a   b   c   d   e   f   g   h
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	2,  4,  4,  4,  4,  4,  4,  2,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  0,  4,  4,  4,  4,  0,  4,
	//
	//	White pawn.
	//
//  a   b   c   d   e   f   g   h
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  1,  2,  2,  1,  0,  0,
	0,  0,  1,  3,  3,  1,  0,  0,
	0,  0,  3,  3,  3,  1,  0,  0,
	0,  0,  3,  3,  3,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
};

//	The king is a weird piece because in the engame its eval is approximately
//	opposite of its eval in the middlegame or opening.  At the start of the
//	game, you want to hide your king in the corner, and at the end, you want
//	to run to the middle.

//	This problem is taken care of by having two tables.  If you try to make
//	things work with the middlegame table, the program will play like crap in
//	endings.

static int	const KingEGPosVal[64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	-3,	0,	0,	0,	0,	0,	0, -3,
	0,	2,	3,	3,	3,	3,	3,	0,
	1,	3,	6,	7,	7,	6,	3,	1,
	1,	3,	7,	8,	8,	7,	3,	1,
	1,	3,	7,	8,	8,	7,	3,	1,
	1,	3,	5,	7,	7,	5,	3,	1,
	0,	2,	4,	3,	3,	4,	2,	0,
	-3, -2,-1,	0,	0,	-1, -2, -3,
};

#define	sideK	0
#define	sideC	1
#define	sideQ	2

#define	sideMAX	3

//	I didn't use the #defines because that is hard to format.  This is an
//	array of char instead of int, which is a decision based upon space.
//
//	Note that c1 is considered a q-side square but f1 is not considered a
//	k-side square.  I want to do everything possible to:
//
//	1.	Convince the program to castle, and O-O-O, which puts the king on
//		c1, is a fine thing to do.
//
//	2.	Prevent the program from playing Kf1.  f1 is not a good square for
//		the king, and I'd rather it play O-O if it's going to move its king
//		into the k-side corner.

static char const c_argside[64] = {
//  a   b   c   d   e   f   g   h
	2,	2,	2,	1,	1,	1,	0,	0,
	2,	2,	1,	1,	1,	1,	0,	0,
	1,	1,	1,	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1,	1,	1,
	2,	2,	1,	1,	1,	1,	0,	0,
	2,	2,	2,	1,	1,	1,	0,	0,
};

//	These are edits to the pawn positional values based upon king location.
//	This one assumes that we're dealing with a white king on the q-side.
//	Non-zero values will replace the values in the regular pawn table.

static char const XXX_argvalShelter[64] = {
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	10,	11,	18,	0,	0,	0,	0,	0,
	15,	15,	18,	0,	0,	0,	0,	0,
	20,	20,	12,	0,	0,	0,	0,	0,
	35,	35,	12,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
};

static int	XXX_argvalPawn[sideMAX][64];

// contain the distance of a square to the border
// used to eval simple mat positions to speedup the finding of the mat
// higher value when opposite king is near the border

const static int c_argdistborder[64] = {
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	10,	10,	10,	10,	10,	10,	0,
	0,	10,	20,	20,	20,	20,	10,	0,
	0,	10,	20,	30,	30,	20,	10,	0,
	0,	10,	20,	30,	30,	20,	10,	0,
	0,	10,	20,	20,	20,	20,	10,	0,
	0,	10,	10,	10,	10,	10,	10,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
};

static int pawn_bonus_to_7[coMAX][8] = {{0,75,45,25,15,5,0,0}};

// f(mat) = mat / 50 if mat > Q+B+K else zero
// all mat => 100; mat==Q+R => 0; mat < Q+R => 0;
static int potentialPressure[32] = {0};
// piece:Q+R+R+B+B+N+N + K
// full score when 8 pieces (start of game)
// 0 or near 0 when only the king left
// mid game position tables are worthless in end game
//#define MidGameValue(n,pop)     (((n)*(pop))/8)
//#define EndGameValue(n,pop)     (((n)*(8-pop))/8)
static int MidGameValueCoef[9] = {0,0,0,0,2,5,7,8,8};
static int EndGameValueCoef[9] = {8,8,8,8,6,2,0,0,0};
#define MidGameValue(n,pop)     (((n)*MidGameValueCoef[pop])/8)
#define EndGameValue(n,pop)     (((n)*EndGameValueCoef[pop])/8)

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

void VInitMiddlegamePawnTable(void)
{
#if 0
	for (int side = sideK; side <= sideQ; side++) {
		for (int co = coWHITE; co <= coBLACK; co++) {
			memcpy(c_argvalPawn[side][co], c_argvalPos[pcPAWN][co],
				csqMAX * sizeof(int));
			if (side != sideC) {
				bool fFlipFil = (side == sideK);
				bool fFlipRnk = (co == coBLACK);

				for (int rnk = rnk2; rnk <= rnk7; rnk++)
					for (int fil = filA; fil <= filH; fil++) {
						int	isq = IsqFromRnkFil(rnk, fil);

						if (c_argvalShelter[isq] != 0) {
							int	rnkEdit = (fFlipRnk) ? 7 - rnk : rnk;
							int	filEdit = (fFlipFil) ? 7 - fil : fil;
							int	isqEdit = IsqFromRnkFil(rnkEdit, filEdit);

							c_argvalPawn[side][co][isqEdit] =
								c_argvalShelter[isq];
						}
					}
			}
		}
	}
#endif
    // TODO:check that ==========================================================
    for(int i = 0; i < 8 ; i++)
        pawn_bonus_to_7[coWHITE][7-i] = pawn_bonus_to_7[coBLACK][i];
    for(int i = 0; i < 32; i++) {
        int valMat = i * 128;
        if( valMat < (valQUEEN))
            potentialPressure[i] = 0;
        else {
            potentialPressure[i] = (128*(valMat-valQUEEN)) / (valROOK+valKNIGHT);
            if(potentialPressure[i] > 128)
                potentialPressure[i] = 128;
        }

    }

}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

static int const c_argdsqKing[64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	-1,	-1,	0,	0,	1,	1,	0,
};

// from Gerd
static unsigned int colorOfSquare (unsigned int sq) {
   return  (0xAA55AA55 >> sq) & 1;
}
// from h.g.m
static bool sameSquareColor(unsigned int sq1, unsigned int sq2) {
    sq1 ^= sq2;
    return  ( (0xAA55AA55 >> sq1) & 1);
}

static int sqDistance( int isqFrom, int isqTo ) {

    int dr = Rank( isqFrom ) - Rank( isqTo );
    int df = File( isqFrom ) - File( isqTo );
    if( dr < 0 ) dr = -dr;
    if( df < 0 ) df = -df;
    if( dr > df )
        return dr;
    return df;
}
// does not check if there is two bishops
bool bishopcanreach(PCON pcon, int side, int sqr ) {
    bitboard bishops = pcon->pieces[BISHOP|side];
    if( bishops == 0 )
        return false;
    int bishop_sqr = LSB( bishops );
    return sameSquareColor(bishop_sqr, sqr);
}

//
// Square rule
// used to see a winning position with KP vs K
// and to give a bonus for passed pawn based on distance to king
//
static bool SquareRuleForPawn(const CON *pcon, const STE *pste, int co, int isqPawn, int & val) {
    // compute ranks from pawn to promotion
    // compute file dist from opposite king to pawn
    // if dist to promote < dist from king then win
    int myturn = pste->side != co;
    int isqKing = pcon->king[co ^ 1];
    int distPawn;
    int rankKing = 7 - Rank( isqKing );
    int rankPawn = 7 - Rank( isqPawn );
    val = 0;
    if( co != coWHITE ) {
        distPawn = 7 - rankPawn;
        rankKing = 7 - rankKing;
    } else
        distPawn = rankPawn;
    if( rankKing > (distPawn + myturn) ) {
        val = 800 - distPawn*64;
        return true;
    }
    int distFil = File( isqKing ) - File( isqPawn );
    if( distFil < 0 )
        distFil = - distFil;
    if( distFil >  (distPawn + myturn) ) {
        val = 800 - distPawn*64;
        return true;
    }
    return false;
}


static bool SquareRuleKPK(PCON pcon, PSTE pste, int co, int & val) {
    // if dist to promote < dist from king then win
    bitboard pawns = pcon->pieces[PAWN|co];
    if(pawns) {
        int isqPawn = PopLSB(&pawns);
        return SquareRuleForPawn(pcon, pste, co, isqPawn, val);
    }
    return false;
}
//	This eval function evaluates:
//
//	1.	Pawn structure, which is hashed and handled in another module.
//	2.	Piece location, which is usually related to central occupation or
//		in the case of pawns, simple advancement.
//	3.	Minors blocking d2/e2 pawns.
//	4.	King on f1/g1 with rook on g1/h1.
//	5.	Pawn shelter.
//	6.	Endgame king position values versus middlegame values.
//
//	Cases 3 and 4 are surprisingly important.
//
//	The piece-square tables are designed to induce the program to castle
//	behind a safe pawn structure, to move its central pawns, and to develop
//	its pieces.
//
//	There is also a chance that the program would feel an interest in breaking
//	its opponent's king position, but I don't have high hopes for this.

int ValEval(PCON pcon, PSTE pste, int Alpha, int Beta)
{
    PHASHPawn phashp;
    int specialVal;
//    bool isEndGame = pste->valPcThem <= 2 * valROOK + valBISHOP;
    bool isPawnEndGame = false;
    int side = pste->side;
    int xside = side ^ 1;
    int isqKingUs = pcon->king[side];
    int isqKingThem = pcon->king[xside];

//    int potentialKingPressure = potentialPressure[ pste->valPcThem / 128 ];
//#define ponderPressure(v)   ((v)*potentialKingPressure)/128

	//	Get value for pawn structure.
	//
    EvalPawns(pcon, pste, phashp);
	//
	//	Get material value.
	//
    int val[2];
    val[side] = pste->valPcUs + pste->valPnUs + phashp->side[side].val ;
    val[xside] = pste->valPcThem + pste->valPnThem + phashp->side[xside].val ;

    //
    //  Check for insuffisant material
    //
    // No pawn
	if ( (pste->valPnUs + pste->valPnThem) == 0 ) {
        // K vs K
		if ( (pste->valPcUs + pste->valPcThem)== 0 )
            return 0;
        // K + Minor vs K
		if ((pste->valPcUs == valBISHOP || pste->valPcUs == valKNIGHT) && (pste->valPcThem == 0))
			return 0;
		if ((pste->valPcUs == 0) && (pste->valPcThem == valBISHOP || pste->valPcThem == valKNIGHT))
			return 0;
        // Two knights can not force checkmate
		if ((pste->valPcUs == 2*valKNIGHT) && (pste->valPcThem == 0))
			return 50;
		if ((pste->valPcUs == 0) && (pste->valPcThem == 2*valKNIGHT))
			return -50;
        // K + Q vs K
		if ((pste->valPcUs == valQUEEN||pste->valPcUs == valROOK) && (pste->valPcThem == 0))
			return pste->valPcUs + 150 - c_argdistborder[ isqKingThem ] -
                sqDistance( isqKingUs, isqKingThem );
        // K vs K + Q | R
		if ((pste->valPcUs == 0) && (pste->valPcThem == valQUEEN||pste->valPcThem == valROOK))
			return -(pste->valPcThem + 150 - c_argdistborder[ isqKingUs ] -
                sqDistance( isqKingUs, isqKingThem ));
    }
    // No piece
    if ((pste->valPcUs + pste->valPcThem)== 0) {
        isPawnEndGame = true;
        // only a pawn on one side
    	if ( (pste->valPnUs == valPAWN) && (pste->valPnThem == 0) ) {
            // compute ranks from pawn to promotion
            // compute file dist from opposite king to pawn
            // if dist to promote < dist from king then win
            if( SquareRuleKPK( pcon, pste, side, specialVal) )
                return specialVal;
        }
    	if ( (pste->valPnUs == 0) && (pste->valPnThem == valPAWN) ) {
            // compute ranks from pawn to promotion
            // compute file dist from opposite king to pawn
            // if dist to promote < dist from king then win
            if( SquareRuleKPK( pcon, pste, xside, specialVal) )
                return -specialVal;
        }
    }
    // bishop + pawns vs nothing
    if ((pste->valPcUs == valBISHOP || pste->valPcUs == 0) && (pste->valPcThem == 0) &&
        (pste->valPnUs > 0) && (pste->valPnThem == 0)) {
        // pawns on file A
        if( (pcon->pieces[PAWN|side] & ~fileAMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, side, promotionSqrFileA[side]) && sqDistance( pcon->king[xside], promotionSqrFileA[side] ) <= 1 )
                return 50;
        // pawns on file H
        if( (pcon->pieces[PAWN|side] & ~fileHMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, side, promotionSqrFileH[side]) && sqDistance( pcon->king[xside], promotionSqrFileH[side] ) <= 1 )
                return 50;
    }
    // bishop + pawns vs nothing
    if ((pste->valPcThem == valBISHOP || pste->valPcThem == 0) && (pste->valPcUs == 0) &&
        (pste->valPnThem > 0) && (pste->valPnUs == 0)) {
        // pawns on file A
        if( (pcon->pieces[PAWN|xside] & ~fileAMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, xside, promotionSqrFileA[xside]) && sqDistance( pcon->king[side], promotionSqrFileA[xside] ) <= 1 )
                return -50;
        // pawns on file H
        if( (pcon->pieces[PAWN|xside] & ~fileHMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, xside, promotionSqrFileH[xside]) && sqDistance( pcon->king[side], promotionSqrFileH[xside] ) <= 1 )
                return -50;
    }


    // lone minor in endgame
    if ((pste->valPcUs == valBISHOP || pste->valPcUs == valKNIGHT) ) {
		val[side] = val[side] - valKNIGHT + 250;
        if (pste->valPnUs == 0)
		    val[side] = val[side] - 100;
    }
    if ((pste->valPcThem == valBISHOP || pste->valPcThem == valKNIGHT) ) {
		val[xside] = val[xside] - valKNIGHT + 250;
        if (pste->valPnThem == 0)
		    val[xside] = val[xside] - 100;
    }

    //
    //	Cut off if it appears that this is futile.
    //  Don't do this cut off in endgames !!
    //
//	if (val - 300 >= Beta)		// We're already great?
//		return Beta;
//	if (val + 300 <= Alpha) 	// We're already toast?
//		return Alpha;
#if 1
    // caution with score and pruning
    int tmpval = val[side] - val[xside];
    int margin = 900;
    if( !isPawnEndGame )
        margin = 300;
	if (tmpval - margin >= Beta)		// We're already great?
        return (tmpval - 50) ? (tmpval - 50) : 1;
	if (tmpval + margin <= Alpha) 	// We're already toast?
        return (tmpval + 50) ? (tmpval + 50) : -1;
#endif

    int	posval[coMAX] = {0,0};  // position value
    int	tacval[coMAX] = {0,0};  // tactical value
    int	otherval[coMAX] = {0,0};  // other value ?
    int	endgval[coMAX] = {0,0};  // end game values
    int notpawn[coMAX] = {0,0};
    int tempoval[coMAX] = {0,0};
    int attackval[coMAX] = {0,0};

    // count of pieces by type and color
    int c_pieces[PIECES] = { 0,0,0,0,0,0,0,0,0,0,0,0 };
    //
    // Rule : first advantage counts full, second and others count half
    //
//    int hasFileAttacker[2] = {popCount(pcon->pieces[ROOK|B]),popCount(pcon->pieces[ROOK|W])};
    _dynamicVal dynamicVal[2];
    computeMobility(pcon, pste, B, dynamicVal[B]);
    computeMobility(pcon, pste, W, dynamicVal[W]);

    // pawns attacked and not defended by other pawns
    // gives a penality because we don't want to defend with a piece
    // with 10 points penality, the engine wants to attack with his queen
    // in the begining
    attackval[W] += 10*popCount( pcon->pieces[PAWN|B] & dynamicVal[W].attacks & ~phashp->side[B].pawnatk );
    attackval[B] += 10*popCount( pcon->pieces[PAWN|W] & dynamicVal[B].attacks & ~phashp->side[W].pawnatk );

    const U8 *pos = pcon->pos;
    // for each piece of the board...
	for (bitboard pieces = pcon->pieces[OCCU]; pieces; ) {
		int sqr = PopLSB(&pieces);
		int pc = pos[sqr];
        int neutralPc = pc & (~1);
        int co = pc & 1;
        // neutral color square for positional values
        int nSqr = remapSq[co][sqr];

        // count kind of pieces
        c_pieces[pc]++;

        // basic positional value
        posval[co] += staticPosVal[neutralPc >> 1][nSqr];
        tempoval[co] += tempoVal[neutralPc >> 1][nSqr];

        if( neutralPc != PAWN )
            notpawn[co]++;

        switch( neutralPc ) {
            case KING:
                if( co == W && Rank(sqr) == 0 || Rank(sqr == 1)) {
                    int f = File(sqr);
                    int prot = 0;
                    if(f > 0 && pos[sqr+7] == WP)   prot ++;
                    if(f < 7 && pos[sqr+9] == WP)   prot ++;
                    if(pos[sqr+8] == WP || pos[sqr+8] == WB)   prot ++;
                    posval[co] += prot * 6;
                }
                if( co == B && Rank(sqr) == 7 || Rank(sqr == 6)) {
                    int f = File(sqr);
                    int prot = 0;
                    if(f > 0 && pos[sqr-9] == BP)   prot ++;
                    if(f < 7 && pos[sqr-7] == BP)   prot ++;
                    if(pos[sqr-8] == BP || pos[sqr-8] == BB)   prot ++;
                    posval[co] += prot * 6;
                }
//                endgval[co] += KingEGPosVal[nSqr];
                break;
            case QUEEN:
                break;
            case ROOK: {
                    int rookF = File(sqr);
                    if(phashp->side[co].c_pawn[rookF] == 0) {
                        posval[co] += 7;
                        if(phashp->side[co^1].c_pawn[rookF] == 0)
                            posval[co] += 7;
                    }
                }
                break;
            case BISHOP:
            case KNIGHT:
                //	Computers are by default very stupid about putting
                //	bishops and knights on d3 and e3 while there is still
                //	a pawn on d2 or e2.  I will detect this case and
                //	penalize heavily.
                //
                if(co == W && (sqr == E3 || sqr == D3) && (pos[sqr - 8] == WP) )
                    otherval[co] -= 20;
                if(co == B && (sqr == E6 || sqr == D6) && (pos[sqr + 8] == BP) )
                    otherval[co] -= 20;

                break;
            case PAWN: {
                bitboard thispawn = (U64)1 << sqr;
                {
                    int rankBonus = pawn_bonus_to_7[co][Rank(sqr)];
                    // give a little bonus for passed pawns
                    if( phashp->side[co].passed & thispawn ) {
                        // this will incite the king to move to candidate/passed pawn of other side
                        endgval[co] += 4*sqDistance(sqr, pcon->king[co^1]);
                        if( isPawnEndGame ) {
                            // compute square rule
                            SquareRuleForPawn(pcon, pste, co, sqr, specialVal);
                            if( specialVal > 0 ) {
                                endgval[co] += (rankBonus + specialVal);
                            } else {
                                endgval[co] += (rankBonus);
                            }
                        } else {
                            posval[co] += 10 + rankBonus / 2;
                            endgval[co] += (rankBonus);
                        }
                    } else if( (~phashp->side[co].blocked) & thispawn ) {
                        // candidate pawn
                        posval[co] += 5 + rankBonus / 4;
                        endgval[co] += (rankBonus);
                        endgval[co] += 4*sqDistance(sqr, pcon->king[co^1]);
                    }
                }
            }
                break;
            default:
                Assert(0);
                break;
        }

	}

    int c_pawn = c_pieces[PAWN] + c_pieces[PAWN|1];

	for (int co = coBLACK; co <= coWHITE; co++) {
        // compute the mobility and position control
        // but only in midgame
        if( MidGameValueCoef[notpawn[ co ^ 1]] > 0 ) {
            // old coefs move * 5/4, offense * 5/4, king * 5
            // mobility has increased a lot now
            // new coefs move * 3/4, offense * 3/4, king * 5
            // 3 is too big
            tacval[co] += (dynamicVal[co].c_move * 2)/4;
            // we must attack the other side so lets give us a bonus for that
            //
            tacval[co] += dynamicVal[co].c_offense * 4/8;
            tacval[co] += dynamicVal[co].c_king * 3;
        }
        if( c_pieces[BISHOP|co] >= 2 ) {
            // bonus is 30 for 0 pawn, 14 for 16 pawn
            posval[co] += (30 - c_pawn)/4;
        }
        if( c_pieces[KNIGHT|co] >= 1 && c_pieces[BISHOP|(co^1)] == 1 )
            posval[co] += (c_pawn - 11)/2;
    }

    // tempoval[co] += tempoVal[neutralPc >> 1][nSqr];

    int dvlp_ratio = notpawn[W] > 4 ? 128 * tempoval[W] / (notpawn[W]*4) : 128;
    if( dvlp_ratio < 85 )
        attackval[W] = attackval[W] * dvlp_ratio / 128;
    dvlp_ratio = notpawn[B] > 4 ? 128 * tempoval[B] / (notpawn[B]*4) : 128;
    if( dvlp_ratio < 85 )
        attackval[B] = attackval[B] * dvlp_ratio / 128;

    int finalVal[2];
	finalVal[side] = val[side] + MidGameValue( 2*tempoval[side] + posval[side] + tacval[side],notpawn[xside]) + 
           EndGameValue( endgval[side], notpawn[xside] ) + otherval[side] + attackval[side];
	finalVal[xside] = val[xside] + MidGameValue( 2*tempoval[xside] + posval[xside] + tacval[xside],notpawn[side]) + 
           EndGameValue( endgval[xside], notpawn[side] ) + otherval[xside] + attackval[xside];

    int delta = finalVal[side] - finalVal[xside];
    int sum = 200 + ((finalVal[side] + finalVal[xside]) > 0) ? (finalVal[side] + finalVal[xside]) : - (finalVal[side] + finalVal[xside]);
    int ratioBonus = delta * 128 / sum;

#ifdef _DEBUG
    if( Alpha == -20000 ) {
        printf("# val=%d ratioBonus=%d => %d (hkey=%I64x)\n", delta, ratioBonus, delta + ratioBonus, pste->hashkPc);
    	for (int co = coBLACK; co <= coWHITE; co++) {
            int dvlp_ratio = notpawn[co] > 4 ? 100 * tempoval[co] / (notpawn[co]*4) : 100;
            printf("# %c val=%d finalval=%d posval=%d tacval=%d endgval=%d other=%d tempo=%d\n",
                "BW"[co], val[co], finalVal[co], posval[co], tacval[co], endgval[co], otherval[co], tempoval[co]);
            printf("# dynamics  dvlp=%d attack=%d move=%d offense=%d king=%d bishops=%d\n", 
                dvlp_ratio, attackval[co], dynamicVal[co].c_move, dynamicVal[co].c_offense, dynamicVal[co].c_king, dynamicVal[co].c_bishop);
        }
    }
#endif
    return (delta ? delta : -2 ) + ratioBonus;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

/*
//  a   b   c   d   e   f   g   h
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	//
	//	White rook.
	//
//  a   b   c   d   e   f   g   h
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	//

*/