//
//
// Copyright (C) 2007  Harald JOHNSEN
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
//
//

#ifndef _EVAL_HXX
#define _EVAL_HXX

// Pawn structure hash element
// The pawn hash contains informations and *static* eval of the pawn structure
// The real value of the pawn structure depends of the complete position (and side
// to move), this value is not integrated in the val tag
// The eval() function will use the pawn structure information to recompute the
// real value depending on the phase of the game

typedef	struct tagHASHP {
	HASHK	hashk;				// Pawn key (other pieces not in here).
//	int	val;					// *static* value from the perspective of white.
    struct tagStructure {
        int c_doubled;          // number of doubled pawn in the position
        int c_islands;          // number of pawn islands
        char c_pawn[8];         // number of pawn per file
        bitboard influence;     // squares that could be attacked by a pawn
        bitboard blocked;       // if pawn is blocked by other side pawn
        bitboard passed;        // not blocked and not within influence(xside)
        bitboard location;      // position of pawns
        bitboard pawnatk;       // squares attack by pawns
        int     val;
    } side[2];                  // this is for each side
    bool pawnon6or7;
}	HASHPawn, * PHASHPawn;

typedef struct {
    // counters
    int c_move;         // non capture + capture + defended material
    int c_king;
    int c_offense;      // number of square in the ennemy camp where we can go
    int c_bishop;       // mobility of bishops
    bitboard    attacks;
} _dynamicVal;

extern void EvalPawns(const CON *pcon, const STE *pste, PHASHPawn & phashp);
extern int const remapSq[2][64];
extern const int exchangePieceValue[16];
extern void computeMobility(CON *pcon, PSTE pste, int side, _dynamicVal & dynamicVal);



#endif // _EVAL_HXX